@@ -1,10 +1,12 @@
 /* $RoughId: sha1init.c,v 1.2 2001/07/13 19:49:10 knu Exp $ */
-/* $Id: sha1init.c 52797 2015-11-30 12:08:16Z usa $ */
+/* $Id: sha1init.c 52694 2015-11-21 04:35:57Z naruse $ */
 
 #include <ruby/ruby.h>
 #include "../digest.h"
-#if defined(HAVE_OPENSSL_SHA_H)
+#if defined(SHA1_USE_OPENSSL)
 #include "sha1ossl.h"
+#elif defined(SHA1_USE_COMMONDIGEST)
+#include "sha1cc.h"
 #else
 #include "sha1.h"
 #endif
@@ -25,7 +27,7 @@ static const rb_digest_metadata_t sha1 = {
  * Technology), described in FIPS PUB 180-1.
  */
 void
-Init_sha1()
+Init_sha1(void)
 {
     VALUE mDigest, cDigest_Base, cDigest_SHA1;
 
@@ -39,6 +41,8 @@ Init_sha1()
 
     cDigest_SHA1 = rb_define_class_under(mDigest, "SHA1", cDigest_Base);
 
-    rb_ivar_set(cDigest_SHA1, rb_intern("metadata"),
-      Data_Wrap_Struct(rb_cObject, 0, 0, (void *)&sha1));
+#undef RUBY_UNTYPED_DATA_WARNING
+#define RUBY_UNTYPED_DATA_WARNING 0
+    rb_iv_set(cDigest_SHA1, "metadata",
+	      Data_Wrap_Struct(0, 0, 0, (void *)&sha1));
 }
